/* Copyright by A.N. Yzelman, Dept. of Mathematics, Utrecht University, 2007-2008.
 * Released under GPL, see the file LICENSE.txt
 */


#include "Triplet.hpp"
#include <vector>

//#define _DEBUG

#ifndef _H_CRS
#define _H_CRS

#ifdef _DEBUG
#include<iostream>
#endif

/**
 *  The compressed row storage sparse matrix data structure.
 */
template< typename T >
class CRS {

  private:

	/** Conveniece typedef */
	typedef unsigned long int ULI;

  protected:

	/** Number of non-zeros. */
	ULI nnz;

	/** Number of rows. */
	ULI nor;

	/** Number of columns */
	ULI noc;

	/** Array keeping track of individual row starting indices. */
	ULI* row_start;

	/** Array containing the actual nnz non-zeros. */
	T* nzs;

	/** Array containing the column indeces corresponding to the elements in nzs. */
	ULI* col_ind;

	/** Which element is considered as zero. */
	T zero_element;

 	/** Sorts 1D columnwise */
        static int compareTriplets( const void * left, const void * right ) {
                const Triplet< T > one = **( (Triplet< T > **)left );
                const Triplet< T > two = **( (Triplet< T > **)right );
                if ( one.j() < two.j() )
                        return -1;
                if ( one.j() > two.j() )
                        return 1;
                return 0;
        }     

	/** 
         *  Helper function which finds a value with a given column index on a given subrange of indices.
         *  @param col_index The given column index.
	 *  @param search_start The start index of the subrange (inclusive).
	 *  @param search_end The end index of the subrange (exlusive).
	 *  @param ret Reference to the variable where the return *index* is stored.
	 *  @return Whether or not a non-zero value should be returned.
	 */
	bool find( const ULI col_index, const ULI search_start, const ULI search_end, ULI &ret ) {
		for( ULI i=search_start; i<search_end; i++ )
			if( col_ind[ i ] == col_index ) {
				ret = i;
				return true;
			}
		return false;
	}

  public:

	/**
	 *  Base constructor which only initialises the internal arrays. Note that to gain a valid CRS structure,
	 *  these arrays have to be filled by some external mechanism (i.e., after calling this constructor, the
	 *  internal arrays contain garbage, resuling in invalid datastructure).
	 *  @param number_of_nonzeros The number of non-zeros to be stored.
	 *  @param number_of_rows The number of rows to be stored.
	 *  @param zero Which element is considered to be the zero element.
	 */
	CRS( const ULI number_of_nonzeros, const ULI number_of_rows, const T zero ):
		nnz( number_of_nonzeros ), nor( number_of_rows ), zero_element( zero ) {
		row_start = new ULI[ nor + 1 ];
		nzs = new T[ nnz ];
		col_ind = new ULI[ nnz ];
	}

	/** Copy constructor.
	 *  @param toCopy reference to the CRS datastructure to copy.
	 */
	CRS( CRS< T >& toCopy ) {
		zero_element = toCopy.zero_element;
		nnz = toCopy.nnz;
		nor = toCopy.nor;
		row_start = new ULI[ nor + 1 ];
		nzs = new T[ nnz ];
		col_ind = new ULI[ nnz ];
		for( ULI i=0; i<nnz; i++ ) {
			nzs[ i ] = toCopy.nzs[ i ];
			col_ind[ i ] = toCopy.col_ind[ i ];
		}
		for( ULI i=0; i<nor; i++ )
			row_start[ i ] = toCopy.row_start[ i ];
	}
	
	/**
	 *  Constructor which transforms a collection of input triplets to CRS format.
	 *  The input collection is considered to have at most one triplet with unique
	 *  pairs of indeces. Unspecified behaviour occurs when this assumption is not
	 *  met.
	 *  @param input The input collection.
	 *  @param m Total number of rows.
	 *  @param n Total number of columns.
	 *  @param zero Which element is considered zero.
	 */
	CRS( std::vector< Triplet< T > > input, unsigned long int m, unsigned long int n, T zero ) {
		zero_element = zero;
		//find nnz
		nnz = input.size();
	
		nor = m;
		noc = n;

		//build better datastructure
		std::vector< std::vector< Triplet< T >* > > ds( nor );
		
		//move input there
		typename std::vector< Triplet< T > >::iterator in_it = input.begin();
		for( ; in_it != input.end(); in_it++ ) {
			Triplet< T >* cur = &(*in_it);
			const ULI currow = cur->i();
			const T value = cur->value;
			if( value == zero_element ) { nnz--; continue; }
			ds.at( currow ).push_back( cur );
		}

		//allocate arrays
		row_start = new ULI[ nor + 1 ];
		nzs = new T[ nnz ];
		col_ind = new ULI[ nnz ];

                //make CRS
                ULI index = 0;
                for( ULI currow = 0; currow < nor; currow++ ) {
                        row_start[ currow ] = index;
                        if( ds.at( currow ).size() == 0 ) continue;
                        qsort( &( ds.at( currow )[ 0 ] ), ds.at( currow ).size(), sizeof( Triplet< T >* ), &compareTriplets );
                        typename std::vector< Triplet< T >* >::iterator row_it = ds.at( currow ).begin();
                        for( ; row_it!=ds.at( currow ).end(); row_it++ ) {
                                const Triplet< T > cur = *(*row_it);
                                nzs[ index ] = cur.value;
                                col_ind[ index ] = cur.j();
                                index++;
                        }
                }
		row_start[ nor ] = nor;
	}


	/**
	 *  Deprecated constructor which transforms a collection of input triplets to CRS format.
	 *  The input collection is considered to have at most one triplet with unique
	 *  pairs of indeces. Unspecified behaviour occurs when this assumption is not
	 *  met.
	 *  Constructor attempt to derive the maximum number of rows, but this is actually a lower bound.
	 *  Constructor does not attempt to find the number of columns; specifically, noc is uninitialised.
	 *  Can be safely used if callee is sure no empty rows exist and does not require this sparse matrix
	 *  structure to perform any operations requireing noc.
	 *  @param input The input collection.
	 *  @param zero Which element is considered zero.
	 */
	CRS( std::vector< Triplet< T > > input, T& zero ) {
		zero_element = zero;
		//find nnz
		nnz = input.size();
	
		//find number of rows
		nor = 0;
		typename std::vector< Triplet< T > >::iterator in_it;
		in_it = input.begin();
		for( ; in_it!=input.end(); in_it++ ) {
			const ULI currow = ( *in_it ).i();
			if ( currow > nor ) nor = currow;
		}
#ifdef _DEBUG
		std::cout << "Max row index found: " << nor << std::endl;
#endif
		nor++;

		//build better datastructure
		std::vector< std::vector< Triplet< T >* > > ds( nor );
		
		//move input there
		in_it = input.begin();
		for( ; in_it != input.end(); in_it++ ) {
			Triplet< T >* cur = &(*in_it);
			const ULI currow = cur->i();
			const T value = cur->value;
			if( value == zero_element ) continue;
			ds.at( currow ).push_back( cur );
		}

		//allocate arrays
		row_start = new ULI[ nor + 1 ];
		nzs = new T[ nnz ];
		col_ind = new ULI[ nnz ];

		//make CRS
		ULI index = 0;
		for( ULI currow = 0; currow < nor; currow++ ) {
#ifdef _DEBUG
			std::cout << "row_start[ " << currow << " ]=" << index << std::endl;
#endif
			row_start[ currow ] = index;
			if( ds.at( currow ).size() == 0 ) continue;
			typename std::vector< Triplet< T >* >::iterator row_it = ds.at( currow ).begin();
			for( ; row_it!=ds.at( currow ).end(); row_it++ ) {
				const Triplet< T > cur = *(*row_it);
				nzs[ index ] = cur.value;
				col_ind[ index ] = cur.j();
				index++;
			}
		}
		row_start[ nor ] = nnz;
	}

	/**
	 *  Method which provides random matrix access to the stored sparse matrix.
	 *  @param i Row index.
	 *  @param j Column index.
	 *  @return Matrix valuei at (i,j).
	 */
	T& random_access( ULI i, ULI j ) {
		ULI found_index;
		if ( find( j, row_start[ i ], row_start[ i+1 ], found_index ) ) {
#ifdef _DEBUG
			std::cout << "Searched col_ind between " << row_start[ i ] << " and " << row_start[ i + 1 ] << ", found: " << std::endl;
			std::cout << "Element (" << i << "," << j << ") found on index " << found_index << ", returning " << nzs[ found_index ] << std::endl;
#endif
			return nzs[ found_index ];
		} else
			return zero_element;
	}

	/**
	 *  Calculates and returns z=Ax. The vectors x should have appropiately set length; this is not
	 *  enforced. Memory leaks, segmentation faults, the works; one or more of these will occur if dimensions
	 *  do not make sense.
	 *  The return array is allocated to length equal to the number of rows in this function.
	 *  @param x The input (dense) x-vector.
	 *  @return The matrix-vector multiplication Ax, where A corresponds to the currently stored matrix.
	 *  @see mv
	 */
	T* mv( T* x ) {
		T* ret = new T[ nor ];
		for( ULI i=0; i<nor; i++ )
			ret[ i ] = zero_element;
		mv( x, ret );
		return ret;
	}

	/** 
	 *  In-place z=Ax function.
	 *  
	 *  @param x The x vector to multiply current matrix with.
	 *  @param z The result vector. Must be pre-allocated and its elements should be initialised to zero.
	 */
        void mv( T* x, T* z ) {
                ULI index=0,row=0;
		for( ; row < nor; row++ )
			for( index = row_start[ row ]; index < row_start[ row + 1 ]; index++ )
				z[ row ] += nzs[ index ] * x[ col_ind[ index ] ];
        }

	/** Returns pointer to the row_start vector. */
	ULI* rowJump() { return row_start; }

	/** Returns pointer to the column index vector. */
	ULI* columnIndices() { return col_ind; }

	/** Returns pointer to the matrix nonzeros vector. */
	T* values() { return nzs; } 

	/** Base deconstructor. */
	~CRS() {
		delete [] row_start;
		delete [] nzs;
		delete [] col_ind;
	}

};

#undef _DEBUG
#endif

