
/* Copyright by A.N. Yzelman, Dept. of Mathematics, Utrecht University, 2007-2008.
 * Released under GPL, see the file LICENSE.txt
 */

#include <vector>
#include "HilbertTriplet.hpp"
#include "Triplet.hpp"

#ifdef _DEBUG
	#include <iostream>
#endif

/** The Hilbert triplet scheme. In effect similar to the triplet scheme (TS),
    but uses Hilbert coordinates to determine the order of storage. */
template< typename T >
class HTS {

   private:
	
	/** Convenience typedef */
	typedef unsigned long int ULI;

   protected:

	/** Minimum number of expansions */
	ULI minexp;

	/** Number of non-zeros. */
	ULI nnz;

	/** Number of rows. */
	ULI nor;

	/** Number of columns. */
	ULI noc;

	/** Vector storing the non-zeros and their locations. */
	std::vector< HilbertTriplet< T > > nzs;

	/** Which element is considered zero. */
	T zero_element;

	/** HilbertCoordinate comparison function. */
	bool cmp( HilbertTriplet< T > &left, HilbertTriplet< T > &right ) {

		if( left.i() == right.i() && left.j() == right.j() ) return true;

		const std::vector< unsigned long int > h_one = left.getHilbert();
		const std::vector< unsigned long int > h_two = right.getHilbert();

		unsigned long int max = h_one.size();
		bool max_is_one = true;
		if ( h_two.size() < max ) { max = h_two.size(); max_is_one = false; }
		for( unsigned long int i=0; i<max; i++ )
			if( h_one[ i ] != h_two[ i ] )
				return h_one[ i ] < h_two[ i ];
#ifdef _DEBUG		
		std::cout << "expand, ";
#endif
		if ( max_is_one )
			left.morePrecision( nor, noc );
		else
			right.morePrecision( nor, noc );

		return cmp( left, right );
	}

	/**
	 *  Binary search for finding a given triplet in a given range.
	 *  @param x triplet to-be found.
	 *  @param left Left bound of the range to search in.
	 *  @param right Right bound of the range to search in.
	 *  @return Index of the triplet searched.
	 */	
	unsigned long int find( HilbertTriplet< T > &x, ULI &left, ULI &right ) {
#ifdef _DEBUG
		std::cout << "left: " << left << ", right: " << right << std::endl;
#endif
		if( nzs[left].getHilbert().size() < nzs[minexp].getHilbert().size() ) minexp = left;
		if( nzs[right].getHilbert().size() < nzs[minexp].getHilbert().size() ) minexp = right;

		if ( left == right ) return left;
		if ( left+1 == right ) return right;
		if ( cmp( x, nzs[ left ] ) ) return left;
		if ( cmp( nzs[ right ], x ) ) return right+1;

		ULI mid = static_cast< unsigned long int >( ( left + right ) / 2 );
#ifdef _DEBUG
		std::cout << "mid: " << mid << std::endl;
#endif
		if ( cmp( x, nzs[ mid ] ) )
			return find( x, left, mid );
		else
			return find( x, mid, right );
	}

   public:

	/** 
	 *  Base constructor.
	 *  Warning: the zero parameter is currently NOT USED!
	 *  @param input Raw input of normal triplets.
	 *  @param m Total number of rows.
	 *  @param n Total number of columns.
	 *  @param zero What elements is considered to-be zero.
	 */
	HTS( std::vector< Triplet< T > > input, ULI m, ULI n, T zero ) {
		zero_element = 0;
		nor = m;
		noc = n;
		nnz = input.size();
		nzs.push_back( HilbertTriplet< T >( input[ 0 ].i(), input[ 0 ].j(), input[ 0 ].value ) );
		minexp = 0;
		for( ULI i=1; i<nnz; i++ ) {
#ifdef _DEBUG
			std::cout << i << ": ";
#endif
			HilbertTriplet< T > h_input = HilbertTriplet< T >( input[i].i(), input[i].j(), input[i].value );
			for( ULI k=0; k<nzs[minexp].getHilbert().size(); k++ ) h_input.morePrecision( nor, noc );
			if( h_input.value != zero_element ) {

				ULI left = 0;
				ULI right = nzs.size() - 1;
				const ULI index = find( h_input, left, right );
#ifdef _DEBUG
				std::cout << "@" << index << ", ";
				std::cout << index << "/" << nzs.size()-1 << std::endl;
#endif
				typename std::vector< HilbertTriplet< T > >::iterator it;

				if( index >= nzs.size() )
					nzs.push_back( h_input );
				else {
					it = nzs.begin();
					it += index;

					nzs.insert( it, h_input );
				}
#ifdef _DEBUG
				std::cout << "inserted; ";
#endif
			}
#ifdef _DEBUG
			std::cout << static_cast< double >( i * 100 ) / static_cast< double >( nnz - 1 ) << "\% Done" << std::endl;
#endif
		}
#ifdef _DEBUG
	        for( ULI i=0; i<nnz; i++ ) {
			for( ULI j=0; j<nzs[i].getHilbert().size(); j++ )
				std::cout << nzs[i].getHilbert()[j] << ",";
			std::cout << std::endl;
		}
#endif
	}

	/** Basic MV. Performs z=Ax.
`	 *  @param x Input vector.
	 *  @return Output vector.
	 */
	T* MV( T* x ) {
		T* ret = new T[ nor ];
		for( ULI i=0; i<nor; i++ )
			ret[ i ] = 0;
		for( ULI i=0; i<nnz; i++ )
			ret[ nzs[i].i() ] += nzs[i].value * x[ nzs[i].j() ];
		return ret;
	}

	/** Saves the current HTS.
	 *  @param fn Filename to save to.
	 *  @see HilbertTriplet< T >::save
`	 */
	void save( const std::string fn ) {
		HilbertTriplet< T >::save( fn, nzs, nor, noc );
	}

};

