
/* Copyright by A.N. Yzelman, Dept. of Mathematics, Utrecht University, 2007-2008.
 * Released under GPL, see the file LICENSE.txt
 */

#include "Triplet.hpp"
#include <assert.h>
#include <vector>
#include <algorithm>

//#define _DEBUG

#ifndef _H_CS_CRS
#define _H_CS_CRS

#ifdef _DEBUG
#include<iostream>
#endif

/**
 *  The *incremental* compressed row storage sparse matrix data structure.
 */
template< typename T >
class ICRS {

  private:

	/** Convenience typedef */
	typedef unsigned long int ULI;

  protected:

	/** Number of non-zeros. */
	ULI nnz;

	/** Number of rows. */
	ULI nor;

	/** Number of cols. */
	ULI noc;

	/** Array containing the actual nnz non-zeros. */
	T* nzs;

	/** Array containing the column jumps. */
	ULI* c_ind;

	/** Array containing the row jumps. */
	ULI* r_ind;

	/** Which element is considered as zero. */
	T zero_element;

	/** Comparison function used for sorting input data. */
	static int compareTriplets( const void * left, const void * right ) {
                const Triplet< T > one = *( (Triplet< T > *)left );
                const Triplet< T > two = *( (Triplet< T > *)right );
                if( one.i() < two.i() )
                        return -1;
                if ( one.i() > two.i() )
                        return 1;
                if ( one.j() < two.j() )
                        return -1;
                if ( one.j() > two.j() )
                        return 1;
                return 0;
        }

  public:

	/**
	 *  Base constructor which only initialises the internal arrays. Note that to gain a valid ICRS structure,
	 *  these arrays have to be filled by some external mechanism (i.e., after calling this constructor, the
	 *  internal arrays contain garbage, resuling in invalid datastructure).
	 *
	 *  @param number_of_nonzeros The number of non-zeros to be stored.
	 *  @param number_of_rows The number of rows of the matrix to be stored.
	 *  @param number_of_cols The number of columns of the matrix to be stored.
	 *  @param zero Which element is considered to be the zero element.
	 */
	ICRS( const ULI number_of_nonzeros, const ULI number_of_rows, const ULI number_of_cols, T zero ):
		nnz( number_of_nonzeros ), noc( number_of_cols ), nor( number_of_rows ), zero_element( zero ) {
		nzs = new T[ nnz ];
		c_ind = new ULI[ nnz ];
		r_ind = new ULI[ nnz ];
	}

	/**
	 *  Copy constructor.
	 *  @param toCopy Reference to the CRS datastructure to copy.
	 */
	ICRS( ICRS< T >& toCopy ) {
		zero_element = toCopy.zero_element;
		nnz = toCopy.nnz;
		noc = toCopy.noc;
		nor = toCopy.nor;
		nzs = new T[ nnz ];
		c_ind = new ULI[ nnz ];
		r_ind = new ULI[ nnz ];
		for( ULI i=0; i<nnz; i = i + 1 ) {
			nzs[ i ] = toCopy.nzs[ i ];
			c_ind[ i ]= toCopy.c_ind[ i ];
			r_ind[ i ] = toCopy.r_ind[ i ];
		}
	}

	/**
	 *  Constructor which transforms a collection of input triplets to CRS format.
	 *  The input collection is considered to have at most one triplet with unique
	 *  pairs of indices. Unspecified behaviour occurs when this assumption is not
	 *  met.
	 *
	 *  @param input The input collection of triplets (i,j,val).
	 *  @param m The number of rows of the input matrix.
	 *  @param n The number of columns of the input matrix.
	 *  @param zero Which element is considered zero.
	 */
	ICRS( std::vector< Triplet< T > >& input, const ULI m, const ULI n, const T zero ) {
		zero_element = zero;

		nor = m;
		noc = n;
	
		typename std::vector< Triplet< T > >::iterator in_it;

		//WARNING: noc*nor typically overflows on 32 bits!
		//         immediately start recording differences
		//         instead of trying to directly calculate
		//         the index as i*noc+j.

		//Complexity compiler package-dependent. Probably O(nnz log(nnz)) average, O(nnz^2) worst-case.
		//for standard C++ sort:
		//sort( input.begin(), input.end(), compareTriplets );
		//for old school C quicksort:
		qsort( &input[ 0 ], input.size(), sizeof( Triplet< T > ), &compareTriplets );

		//filtering out zeros is skipped for now.
		
		//Count the number of row jumps
		std::vector< ULI > r_ind_temp;
		typename std::vector< Triplet< T > >::iterator it = input.begin();
		ULI prev_row = (*it).i();
		r_ind_temp.push_back( prev_row );
		it++;

		//O(nnz log(nnz)); push_back on a vector uses amortised log(n) array growing algorithm on inserts
		for( ; it!=input.end(); it++ ) {
			if( (*it).i() > prev_row ) {
				r_ind_temp.push_back( (*it).i() - prev_row );
				prev_row = (*it).i();
			}
		}
		nnz = input.size();

		//allocate arrays
		nzs = new T[ nnz ];
		c_ind = new ULI[ nnz + 1 ];
		r_ind = new ULI[ r_ind_temp.size() + 1 ];

		//set last entry
		c_ind[ nnz ] = noc;
		//r_ind does not have to be set; altough the last element is read, it is actually never used.

		//copy row-jump vector
		//O(m) worst case
		for( ULI i=0; i<r_ind_temp.size(); i++ )
			r_ind[ i ] = r_ind_temp[ i ];
		//clear temporary r_ind vector
		r_ind_temp.clear();

		//make ICRS
		ULI prev_col = 0;
		prev_row = 0;

		//O(nnz)
		for( ULI i=0; i<nnz; ++i ) {
			const Triplet< T > cur = input[ i ];
			const ULI currow = cur.i();
			const ULI curcol = cur.j();
			if( currow == prev_row ) {
				c_ind[ i ] = curcol - prev_col;
			} else {
				c_ind[ i ] = noc + ( curcol - prev_col );
				prev_row = currow;
			}
			nzs[ i ] = cur.value;
			prev_col = curcol;

#ifdef _DEBUG
			std::cout << currow << "," << curcol << "(" << cur.value << ") maps to " << c_ind[ i ] << std::endl;
#endif
		}
	}

	/*
	 *  Method which provides random matrix access to the stored sparse matrix.
	 *  @param i Row index.
	 *  @param j Column index.
	 *  @return Matrix value at (i,j).
	 */
	/*T& random_access( const ULI& i, const ULI& j ) {
		return zero_element; //TODO
	}*/

	/**
	 *  Calculates and returns z=Ax. The vector x should have appropiately set length; this is not strictly
	 *  enforced. Bus errors, segmentation faults, the works; one or more of these will occur if dimensions
	 *  do not make sense.
	 *  The return array is allocated to length equal to the number of rows in this function.
	 *  @param x The input (dense) x-vector.
	 *  @return The matrix-vector multiplication Ax, where A corresponds to the currently stored matrix.
	 */
	T* mv( T* x ) {
 		T* ret = new T[ nor ];
		for( ULI i=0; i < nor; ++i )
			ret[ i ] = zero_element;
		mv( x, ret );
		return ret;
	}

	/**
	 *  In-place z=Ax function. Adapted from the master thesis of Joris Koster, Utrecht University.
	 *
	 *  @param pDataX Pointer to array x to multiply by the current matrix (Ax).
	 *  @param pDataZ Pointer to result array. Must be pre-allocated and its elements set to zero for correct results.
	 */
        void mv( T*__restrict__ pDataX, T*__restrict__ pDataZ ) {
		         T *__restrict__ pDataA    = nzs;
		const    T *__restrict__ pDataAend = nzs + nnz;
		//register T *__restrict__ pDataX    = x; //use of register keyword is discouraged
		const    T *__restrict__ pDataXend = pDataX + noc;
		//register T *__restrict__ pDataZ    = z; //use of register keyword is discouraged
		//const    T *pDataZend = z + nor; //unused

		//register ULI *__restrict__ pIncRow   = r_ind; //use of register keyword is discouraged
		//register ULI *__restrict__ pIncCol   = c_ind; //use of register keyword is discouraged
		ULI *__restrict__ pIncRow   = r_ind;
		ULI *__restrict__ pIncCol   = c_ind;

		//go to first column
		pDataX += *pIncCol++;
		while( pDataA < pDataAend ) {
			//jump to correct row
			pDataZ += *pIncRow;
			while( pDataX < pDataXend ) {
				*pDataZ += *pDataA * *pDataX;
				pDataA   = pDataA + 1;
				pDataX  += *pIncCol;
				pIncCol  = pIncCol + 1;
			}
			pDataX -= noc;
			pIncRow = pIncRow + 1;
		}
        }

	/** Base deconstructor. */
	~ICRS() {
		delete [] nzs;
		delete [] c_ind;
		delete [] r_ind;
	}

};

#endif

