
/* Copyright by A.N. Yzelman, Dept. of Mathematics, Utrecht University, 2007-2008.
 * Released under GPL, see the file LICENSE.txt
 */

#include <vector>
#include <assert.h>
#include "Triplet.hpp"

/** The triplet scheme; a storage scheme for sparse matrices using triplets. */
template< typename T >
class TS {

   private:
	
	/** Convenience typedef */
	typedef unsigned long int ULI;

   protected:

	/** Number of non-zeros. */
	ULI nnz;

	/** Number of rows. */
	ULI nor;

	/** Number of columns. */
	ULI noc;

	/** The row indices of the nonzeros. */
	ULI* i;
	
	/** The column indices of the nonzeros. */
	ULI* j;

	/** The values of the nonzeros. */
	T* nzs;

	/** Which element is considered zero. */
	T zero_element;

   public:

	/** Base constructor.
	 *  @param input std::vector of triplets to be stored in this scheme.
	 *  @param m total number of rows.
	 *  @param n total number of columns.
	 *  @param zero what is to be considered the zero element.
	 */
	TS( std::vector< Triplet< T > > input, ULI m, ULI n, T zero ) {
		ULI offset = 0;

		zero_element = zero;
		nor = m;
		noc = n;
		nnz = input.size();
		nzs = new T[ nnz ];
		i = new ULI[ nnz ];
		j = new ULI[ nnz ];
		for( ULI r=0; r<nnz; r++ )
			if( input[ r ].value != zero_element ) {
				//nzs[ r ] = input[ r ];
				nzs[ r - offset ] = input[ r ].value;
				i[ r - offset ] = input[ r ].i();
				j[ r - offset ] = input[ r ].j();
			} else {
				offset++;
			}
		nnz -= offset;
	}

	/** Basic MV. Performs z=Ax.
	 *  @param x input vector.
	 *  @return output vector.
	 */
	T* mv( T* x ) {
		T* ret = new T[ nor ];
		for( ULI r=0; r<nor; r++ )
			ret[ r ] = zero_element;
		for( ULI r=0; r<nnz; r++ ) {
			assert( i[ r ] >= 0  );
			assert( i[ r ] < nor );
                        assert( j[ r ] >= 0  );
			assert( j[ r ] < noc );
			ret[ i[ r ] ] += nzs[ r ] * x[ j[ r ] ];
		}
		return ret;
	}

	/**
	 *  In-place z=Ax calculation algorithm.
	 *
	 *  @param x The vector x supplied for calculation of Ax.
	 *  @param z The result vector z. Should be pre-allocated with entries set to zero.
	 */
        void mv( T* x, T* z ) {
                for( ULI r=0; r<nnz; r++ ) {
                        assert( i[ r ] >= 0  );
                        assert( i[ r ] < nor );
                        assert( j[ r ] >= 0  );
                        assert( j[ r ] < noc );
                        z[ i[ r ] ] += nzs[ r ] * x[ j[ r ] ];
                }
        }

	/** Base destructor. */
	~TS() {
		delete [] i;
		delete [] j;
		delete [] nzs;
	}

};


