
/* Copyright by A.N. Yzelman, Dept. of Mathematics, Utrecht University, 2007-2008.
 * Released under GPL, see the file LICENSE.txt
 */

#include "Triplet.hpp"
#include <assert.h>
#include <vector>
#include <algorithm>

//#define _DEBUG

#ifndef _H_ZZ_ICRS
#define _H_ZZ_ICRS

//#define _UNSIGNED

//#ifdef _DEBUG
#include<iostream>
//#endif

/**
 *  The zig-zag incremental compressed row storage sparse matrix data structure.
 */
template< typename T >
class ZZ_ICRS {

  private:

#ifdef _UNSIGNED
	/** Conveniece typedef */
	typedef unsigned long int ULI;
#else
	/**Convenience typedef */
	typedef long int ULI;
#endif

  protected:

	/** Number of non-zeros. */
	ULI nnz;

	/** Number of rows. */
	ULI nor;

	/** Number of cols. */
	ULI noc;

	/** Array containing the actual nnz non-zeros. */
	T* nzs;

	/** Array containing the column jumps. */
	ULI* c_ind;

	/** Array containing the row jumps. */
	ULI* r_ind;

	/** Which element is considered as zero. */
	T zero_element;

	/** Comparison function used for sorting input data. */
	static bool compareTriplets( const Triplet< T >& one, const Triplet< T >& two ) {
		if( one.i() < two.i() )
			return true;
		if ( one.i() > two.i() )
			return false;
		return one.j() < two.j();
	}

  public:

	/**
	 *  Base constructor which only initialises the internal arrays. Note that to gain a valid CRS structure,
	 *  these arrays have to be filled by some external mechanism (i.e., after calling this constructor, the
	 *  internal arrays contain garbage, resuling in invalid datastructure).
	 *  @param number_of_nonzeros The number of non-zeros to be stored.
	 *  @param number_of_rows The number of rows to be stored.
	 *  @param number_of_cols The number of columns to be stored.
	 *  @param zero Which element is considered to be the zero element.
	 */
	ZZ_ICRS( const ULI number_of_nonzeros, const ULI number_of_rows, const ULI number_of_cols, T zero ):
		nnz( number_of_nonzeros ), noc( number_of_cols ), nor( number_of_rows ), zero_element( zero ) {
		nzs = new T[ nnz ];
		c_ind = new ULI[ nnz ];
		r_ind = new ULI[ nnz ];
	}

	/**
	 *  Copy constructor.
	 *  @param toCopy reference to the CRS datastructure to copy.
	 */
	ZZ_ICRS( ZZ_ICRS< T >& toCopy ) {
		zero_element = toCopy.zero_element;
		nnz = toCopy.nnz;
		noc = toCopy.noc;
		nor = toCopy.nor;
		nzs = new T[ nnz ];
		c_ind = new ULI[ nnz ];
		r_ind = new ULI[ nnz ];
		for( ULI i=0; i<nnz; i = i + 1 ) {
			nzs[ i ] = toCopy.nzs[ i ];
			c_ind[ i ]= toCopy.c_ind[ i ];
			r_ind[ i ] = toCopy.r_ind[ i ];
		}
	}

	/**
	 *  Constructor which transforms a collection of input triplets to CRS format.
	 *  The input collection is considered to have at most one triplet with unique
	 *  pairs of indeces. Unspecified behaviour occurs when this assumption is not
	 *  met.
	 *  @param input The input collection.
	 *  @param m Total number of rows.
	 *  @param n Total number of columns.
	 *  @param zero Which element is considered zero.
	 */
	ZZ_ICRS( std::vector< Triplet< T > >& input, const ULI m, const ULI n, const T zero ) {
		zero_element = zero;

		nor = m;
		noc = n;
	
		typename std::vector< Triplet< T > >::iterator in_it;

		//WARNING: noc*nor typically overflows on 32 bits!
		//         immediately start recording differences
		//         instead of trying to directly calculate
		//         the index as i*noc+j.
	
		sort( input.begin(), input.end(), compareTriplets );

		//filter out zeros and count the number of row jumps
		std::vector< ULI > r_ind_temp;
		typename std::vector< Triplet< T > >::iterator it = input.begin();
		unsigned long int prev_row = (*it).i();
		r_ind_temp.push_back( prev_row );
		it++;
		for( ; it!=input.end(); it++ ) {
			if( (*it).i() > prev_row ) {
				r_ind_temp.push_back( (*it).i() - prev_row );	
				prev_row = (*it).i();
			}
		}
		nnz = input.size();

		//allocate arrays
		nzs = new T[ nnz ];
		c_ind = new ULI[ nnz + 1 ];
		r_ind = new ULI[ r_ind_temp.size() + 1 ];

		//set last entry
		c_ind[ nnz ] = noc;
		//r_ind does not have to be set; altough the last element is read, it is actually never used.

		//copy row-jump vector
		for( ULI i=0; i<static_cast< ULI >( r_ind_temp.size() ); i++ )
			r_ind[ i ] = r_ind_temp[ i ];

		//make ICRS
		ULI prev_col = 0;
		prev_row = 0;

		for( ULI i=0; i<nnz; ++i ) {
			const Triplet< T > cur = input[ i ];
			const unsigned long int currow = cur.i();
			const unsigned long int curcol = cur.j();
			if( currow == prev_row ) {
				c_ind[ i ] = curcol - prev_col;
				nzs[ i ] = cur.value;
				prev_col = curcol;
			} else {
				prev_row = currow;
				ULI stop = i;
				ULI j = i;
				
				while( j < nnz && input[ j ].i() == static_cast< unsigned long int >( currow ) )
					j++; //go to next row
				//ULI next = j; //where the next row starts, OR next = j = nnz // unused
				j--; // j now points at last element at new row

				//process the first new element
				nzs[ i ] = input[ j ].value;
				c_ind[ i ] = noc + ( input[ j ].j() - prev_col );
				prev_col = input[ j ].j();

				i++; //go to next indices
				j--;

				for( ; j>=stop; j-- ) { //reversively add the rest to data structure
					nzs[ i ] = input[ j ].value;
					c_ind[ i ] = prev_col - input[ j ].j();
					prev_col = input[ j ].j();
					i++; //increment target index
				}

				if( i >= nnz ) break;

				//at this point, the reverse row is completely added
				//and i should be at the next row index.
#ifdef _UNSIGNED
				c_ind[ i ] = noc + input[ i ].j();
#else
				c_ind[ i ] = noc + ( prev_col - input[ i ].j() );
#endif

				nzs[ i ] = input[ i ].value;
				prev_col = input[ i ].j();
				prev_row = input[ i ].i();
			}
		}
	}

	/**
	 *  Method which provides random matrix access to the stored sparse matrix.
	 *  Note: currently, use of the STL algorithm 'find' is employed. This might
	 *        very well *not* be the most efficient algorithm used on a sorted
	 *        datastructure as ICRS.
	 *
	 *  @param i Row index.
	 *  @param j Column index.
	 *  @return Matrix value at (i,j).
	 */
	/*T& random_access( const ULI& i, const ULI& j ) {
		return zero_element; //TODO
	}*/

	/**
	 *  Calculates and returns z=Ax. The vector x should have appropiately set length; this is not strictly
	 *  enforced. Bus errors, segmentation faults, the works; one or more of these will occur if dimensions
	 *  do not make sense.
	 *  The return array is allocated to length equal to the number of rows in this function.
	 *  @param x The input (dense) x-vector.
	 *  @return The matrix-vector multiplication Ax, where A corresponds to the currently stored matrix.
	 */
	T* mv( T* x ) {
		T* ret = new T[ nor ];
		for( ULI i=0; i < nor; ++i )
			ret[ i ] = zero_element;
		mv( x, ret );
		return ret;
	}

	/**
	 *  In-place z=Ax multiplication algorithm.
	 *
	 *  @param x The vector x supplied for multiplication.
	 *  @param z The pre-allocated result vector. All elements should be set to zero in advance.
	 */
	void mv( T* x, T* z ) {

		//get some pointers
		T   *pDataA    = nzs;
		T   *pDataAend = nzs + nnz;
		T   *pDataXbeg = x;
		T   *pDataX    = x;
		T   *pDataXend = x + noc;
		T   *pDataZ    = z;
		ULI *pIncRow   = r_ind;
		ULI *pIncCol   = c_ind;

		ULI rowshadow  = 0;

		//go to first column
		pDataX += *pIncCol;
		pIncCol++;

		//walk through all matrix entries
		while( pDataA < pDataAend ) {

			//jump to correct row
			pDataZ += *pIncRow;
			rowshadow += *pIncRow;

			//walk left-to-right
			while( pDataX < pDataXend ) {

				*pDataZ  += *pDataA  * *pDataX;
				 pDataA   =  pDataA  + 1      ;
				 pDataX  += *pIncCol          ;
				 pIncCol  =  pIncCol + 1      ;

			}

			//check if there is no more data
			if( pDataA >= pDataAend ) break;

			//jump to correct row and column
			pDataX  -=  noc        ;
			pIncRow  =  pIncRow + 1;
			pDataZ  += *pIncRow    ;
			rowshadow+=*pIncRow;
			assert( rowshadow >= 0  );
			assert( rowshadow < nor );

			//walk right-to-left
			while( pDataXbeg <= pDataX ) {
				
				*pDataZ += *pDataA  * *pDataX;
				 pDataA  =  pDataA  + 1      ;
				 pDataX -= *pIncCol          ;
 				 pIncCol =  pIncCol + 1      ;

			}

			//jump to crrect column
			pDataX += noc;

			//set row increment pointer to crrect location
			pIncRow = pIncRow + 1;
			
		}

	}

	/** Base deconstructor. */
	~ZZ_ICRS() {
		delete [] nzs;
		delete [] c_ind;
		delete [] r_ind;
	}

};

#endif

